from ANNarchy.core import Global
from .SharedProjection import SharedProjection

Global._warning("ANNarchy.extensions.weightsharing is deprecated since 4.6.8, use ANNarchy.extensions.convolution instead.")
